@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2014 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -57,16 +57,17 @@
 #include "wrapperinfo.h"
 #include "property.h"
 #include "logger.h"
+#include "wrapper_file.h"
 
 #include <sys/resource.h>
 #include <sys/time.h>
 
 #if defined(IRIX)
-#define PATH_MAX FILENAME_MAX
+ #define PATH_MAX FILENAME_MAX
 #endif
 
 #ifndef USE_USLEEP
-#include <time.h>
+ #include <time.h>
 #endif
 
 #ifndef getsid
@@ -154,27 +155,22 @@ void appExit(int exitCode, int argc, TCHAR** argv) {
     exit(exitCode);
 }
 
-/**
- * Gets the error code for the last operation that failed.
- */
-int wrapperGetLastError() {
-    return errno;
-}
-
 /**
  * Writes a PID to disk.
  *
  * filename: File to write to.
  * pid: pid to write in the file.
+ * strict: If true then an error will be reported and the call will fail if the
+ *         file already exists.
+ *
+ * return 1 if there was an error, 0 if Ok.
  */
-int writePidFile(const TCHAR *filename, DWORD pid, int newUmask, int exclusive) {
+int writePidFile(const TCHAR *filename, DWORD pid, int newUmask, int strict) {
     FILE *pid_fp = NULL;
     int old_umask;
 
-    if (getBooleanProperty(properties, TEXT("wrapper.pidfile.strict"), FALSE, FALSE) == TRUE && 
-        exclusive == TRUE && file_exists(filename)) {
-        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
-            TEXT("%d pid file, %s, already exists."), pid, filename);
+    if (strict && wrapperFileExists(filename)) {
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR, TEXT("%d pid file, %s, already exists."), pid, filename);
         return 1;
     }
 
@@ -235,13 +231,18 @@ void takeSignalAction(int sigNum, const TCHAR *sigName, int mode) {
                 (wrapperData->jState == WRAPPER_JSTATE_DOWN_FLUSH)) {
 
                 /* Signaled while we were already shutting down. */
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("%s trapped.  Forcing immediate shutdown."), sigName);
-
-                /* Disable the thread dump on exit feature if it is set because it
-                 *  should not be displayed when the user requested the immediate exit. */
-                wrapperData->requestThreadDumpOnFailedJVMExit = FALSE;
-                wrapperKillProcess();
+                if (wrapperData->isForcedShutdownDisabled) {
+                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
+                        TEXT("%s trapped.  Already shutting down."), sigName);
+                } else {
+                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
+                        TEXT("%s trapped.  Forcing immediate shutdown."), sigName);
+    
+                    /* Disable the thread dump on exit feature if it is set because it
+                     *  should not be displayed when the user requested the immediate exit. */
+                    wrapperData->requestThreadDumpOnFailedJVMExit = FALSE;
+                    wrapperKillProcess();
+                }
             } else {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
                     TEXT("%s trapped.  Shutting down."), sigName);
@@ -468,13 +469,17 @@ void descSignal(siginfo_t *sigInfo) {
  #ifndef UNICODE
                 uName = pw->pw_name;
  #else
-                req = mbstowcs(NULL, pw->pw_name, 0) + 1;
-                uName = malloc(req * sizeof(TCHAR));
+                req = mbstowcs(NULL, pw->pw_name, MBSTOWCS_QUERY_LENGTH);
+                if (req == (size_t)-1) {
+                    return;
+                }
+                uName = malloc(sizeof(TCHAR) * (req + 1));
                 if (!uName) {
                     outOfMemory(TEXT("DSCS"), 1);
                     return;
                 }
                 mbstowcs(uName, pw->pw_name, req + 1);
+                uName[req] = TEXT('\0'); /* Avoid bufferflows caused by badly encoded characters. */
  #endif
             }
 
@@ -694,8 +699,8 @@ void *javaIORunner(void *arg) {
     /* Loop until we are shutting down, but continue as long as there is more output from the JVM. */
     while ((!stopJavaIOThread) || (!nextSleep)) {
         if (nextSleep) {
-            /* Sleep for a hundredth of a second. */
-            wrapperSleep(10);
+            /* Sleep as little as possible. */
+            wrapperSleep(1);
         }
         nextSleep = TRUE;
         
@@ -704,7 +709,7 @@ void *javaIORunner(void *arg) {
             wrapperData->pauseThreadJavaIO = 0;
         }
         
-        if (wrapperReadChildOutput()) {
+        if (wrapperReadChildOutput(0)) {
             if (wrapperData->isDebugging) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_DEBUG,
                     TEXT("Pause reading child process output to share cycles."));
@@ -761,7 +766,7 @@ void disposeJavaIO() {
 #endif
             wrapperSleep(100);
         }
-        pthread_kill(javaIOThreadId, SIGKILL);
+        pthread_cancel(javaIOThreadId);
     }
 }
 
@@ -911,7 +916,7 @@ void disposeTimer() {
 #endif
             wrapperSleep(100);
         }
-        pthread_kill(timerThreadId, SIGKILL);
+        pthread_cancel(timerThreadId);
     }
 }
 
@@ -1148,6 +1153,10 @@ void wrapperExecute() {
      *  threads do not reopen the log file as the new process is being created. */
     setLogfileAutoClose(TRUE);
     closeLogfile();
+        
+    /* Reset the log duration so we get new counts from the time the JVM is launched. */
+    resetDuration();
+    
     /* Fork off the child. */
     proc = fork();
 
@@ -1155,7 +1164,7 @@ void wrapperExecute() {
         /* Fork failed. */
 
         /* Restore the auto close flag. */
-        setLogfileAutoClose(wrapperData->logfileInactivityTimeout <= 0);
+        setLogfileAutoClose(wrapperData->logfileCloseTimeout == 0);
 
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
                    TEXT("Could not spawn JVM process: %s"), getLastErrorText());
@@ -1201,6 +1210,19 @@ void wrapperExecute() {
             close(pipedes[PIPE_WRITE_END]);
             pipedes[PIPE_WRITE_END] = -1;
 
+            /* forking at this point, the child process has set all pipes already, so no
+               further assignments needed */
+            if (wrapperData->printJVMVersion) {
+                if (fork() == 0) {
+                    /* in the child */
+                    TCHAR *javaVersionArgv[3];
+                    javaVersionArgv[0] = wrapperData->jvmCommand[0];
+                    javaVersionArgv[1] = TEXT("-version");
+                    javaVersionArgv[2] = 0;
+                    _texecvp(wrapperData->jvmCommand[0], javaVersionArgv);
+                }
+                /* we don't care about errors here, since they will be reported later */
+            }
             /* The pipedes array is global so do not close the pipes. */
             /* Child process: execute the JVM. */
             _texecvp(wrapperData->jvmCommand[0], wrapperData->jvmCommand);
@@ -1241,7 +1263,7 @@ void wrapperExecute() {
             pipedes[PIPE_WRITE_END] = -1;
 
             /* Restore the auto close flag. */
-            setLogfileAutoClose(wrapperData->logfileInactivityTimeout <= 0);
+            setLogfileAutoClose(wrapperData->logfileCloseTimeout == 0);			
 
             /* The pipedes array is global so do not close the pipes. */
 
@@ -1266,7 +1288,7 @@ void wrapperExecute() {
                 }
             }
 
-            /* If a java id filename is specified then write the pid of the java process. */
+            /* If a java id filename is specified then write the Id of the java process. */
             if (wrapperData->javaIdFilename) {
                 if (writePidFile(wrapperData->javaIdFilename, wrapperData->jvmRestarts, wrapperData->javaIdFileUmask, FALSE)) {
                     log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
@@ -1605,7 +1627,7 @@ void daemonize(int argc, TCHAR** argv) {
     }
 
     /* Restore the auto close flag in the daemonized process. */
-    setLogfileAutoClose(wrapperData->logfileInactivityTimeout <= 0);
+    setLogfileAutoClose(wrapperData->logfileCloseTimeout == 0);
 }
 
 
@@ -1613,11 +1635,11 @@ void daemonize(int argc, TCHAR** argv) {
  * Sets the working directory to that of the current executable
  */
 int setWorkingDir(TCHAR *app) {
-    TCHAR szPath[PATH_MAX + 1];
+    TCHAR *szPath;
     TCHAR* pos;
 
     /* Get the full path and filename of this program */
-    if (_trealpath(app, szPath) == NULL) {
+    if ((szPath = findPathOf(app, TEXT("Wrapper binary"))) == NULL) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("Unable to get the path for '%s'-%s"),
             app, getLastErrorText());
         return 1;
@@ -1632,6 +1654,7 @@ int setWorkingDir(TCHAR *app) {
     pos = _tcsrchr(szPath, TEXT('/'));
     if (pos == NULL) {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL, TEXT("Unable to extract path from: %s"), szPath);
+        free(szPath);
         return 1;
     } else {
         /* Clip the path at the position of the last backslash */
@@ -1641,10 +1664,11 @@ int setWorkingDir(TCHAR *app) {
     /* Set a variable to the location of the binary. */
     setEnv(TEXT("WRAPPER_BIN_DIR"), szPath, ENV_SOURCE_WRAPPER);
 
-    if (wrapperSetWorkingDir(szPath)) {
+    if (wrapperSetWorkingDir(szPath, TRUE)) {
+        free(szPath);
         return 1;
     }
-
+    free(szPath);
     return 0;
 }
 
@@ -1664,6 +1688,15 @@ int main(int argc, char **argv) {
 #ifdef UNICODE
     size_t req;
     TCHAR **argv;
+
+#ifdef FREEBSD
+    /* In the case of FreeBSD, we need to dynamically load and initialize the iconv library to work with all versions of FreeBSD. */
+    if (loadIconvLibrary()) {
+        /* Already reported. */
+        /* Don't call appExit here as we are not far enough along. */
+        return 1;
+    }
+#endif    
     
     /* Create UNICODE versions of the argv array for internal use. */
     argv = malloc(argc * sizeof *argv );
@@ -1673,8 +1706,12 @@ int main(int argc, char **argv) {
         return 1;
     }
     for (i = 0; i < argc; i++) {
-        req = mbstowcs(NULL, cargv[i], 0);
-        argv[i] = malloc((int)(req + 1) * sizeof(TCHAR));
+        req = mbstowcs(NULL, cargv[i], MBSTOWCS_QUERY_LENGTH);
+        if (req == (size_t)-1) {
+            _tprintf(TEXT("Encoding problem with arguments in Main\n"));
+            return 1;
+        }
+        argv[i] = malloc(sizeof(TCHAR) * (req + 1));
         if (!argv[i]) {
             _tprintf(TEXT("Out of Memory in Main\n"));
             while(--i > 0) {
@@ -1684,7 +1721,8 @@ int main(int argc, char **argv) {
             appExit(1, 0, argv);
             return 1;
         }
-        mbstowcs(argv[i], cargv[i], (req + 1) * sizeof(TCHAR));
+        mbstowcs(argv[i], cargv[i], req + 1);
+        argv[i][req] = TEXT('\0'); /* Avoid bufferflows caused by badly encoded characters. */
     }
 
 #endif
@@ -1787,6 +1825,9 @@ int main(int argc, char **argv) {
         /* fork to a Daemonized process if configured to do so. */
         if (wrapperData->daemonize) {
             daemonize(argc, argv);
+            
+            /* We are now daemonized, so mark this as being a service. */
+            wrapperData->isConsole = FALSE;
 
             /* When we daemonize the Wrapper, its PID changes. Because of the
              *  WRAPPER_PID environment variable, we need to set it again here
@@ -1804,7 +1845,7 @@ int main(int argc, char **argv) {
              *  the configuration can be reloaded.  This is needed to support relative
              *  references to include files. */
             if (wrapperData->workingDir && wrapperData->originalWorkingDir) {
-                if (wrapperSetWorkingDir(wrapperData->originalWorkingDir)) {
+                if (wrapperSetWorkingDir(wrapperData->originalWorkingDir, TRUE)) {
                     /* Failed to restore the working dir.  Shutdown the Wrapper */
                     appExit(1, argc, argv);
                     return 1; /* For compiler. */
@@ -1833,7 +1874,7 @@ int main(int argc, char **argv) {
         }
 
         if (wrapperData->pidFilename) {
-            if (writePidFile(wrapperData->pidFilename, wrapperData->wrapperPID, wrapperData->pidFileUmask, TRUE)) {
+            if (writePidFile(wrapperData->pidFilename, wrapperData->wrapperPID, wrapperData->pidFileUmask, wrapperData->pidFileStrict)) {
                 log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_FATAL,
                      TEXT("ERROR: Could not write pid file %s: %s"),
                 wrapperData->pidFilename, getLastErrorText());
@@ -1878,7 +1919,11 @@ int main(int argc, char **argv) {
             }
         }
 
-        appExit(wrapperRunConsole(), argc, argv);
+        if (wrapperData->isConsole) {
+            appExit(wrapperRunConsole(), argc, argv);
+        } else {
+            appExit(wrapperRunService(), argc, argv);
+        }
         return 0; /* For compiler. */
     } else {
         log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN, TEXT(""));
